use std::collections::HashMap;
use std::fmt::Write;
use std::io;
use std::os::unix::prelude::*;
use std::path::PathBuf;
use std::str;
use std::sync::{
    atomic::{AtomicBool, AtomicUsize, Ordering},
    Arc,
};
use std::time::Duration;

use aho_corasick::AhoCorasick;
use clap::Parser;
use dashmap::DashMap;
use indicatif::{ProgressBar, ProgressState, ProgressStyle};
use once_cell::sync::Lazy;
use tokio::{fs, sync::Semaphore, time::sleep};
use tracing::Instrument;

static KEYWORDS: &[&str] = &[
    "数字经济",
    "数字媒体",
    "数字化",
    "数字资产",
    "数字化转型",
    "数据管理",
    "数据挖掘",
    "数据网络",
    "数据平台",
    "数据中心",
    "数据科学",
    "数字控制",
    "数字技术",
    "数字通信",
    "数字网络",
    "数字智能",
    "数字终端",
    "数字营销",
    "大数据",
    "云计算",
    "云IT",
    "云生态",
    "云服务",
    "云平台",
    "区块链",
    "物联网",
    "机器学习",
];
static ANALYZER: Lazy<AhoCorasick> = Lazy::new(|| AhoCorasick::new(KEYWORDS.iter()).unwrap());

#[derive(Parser, Debug)]
#[command(version, about, long_about = None)]
struct Args {
    #[arg(short, long)]
    directory: PathBuf,
    #[arg(short, long, default_value = "16")]
    jobs: usize,
}

#[tokio::main]
async fn main() -> io::Result<()> {
    tracing_subscriber::fmt::init();

    let args = Args::parse();
    tracing::debug!(
        "start to analyze {} with {} jobs",
        args.directory.display(),
        args.jobs
    );

    let all_tasks = {
        let mut res = Vec::new();
        let mut files = fs::read_dir(&args.directory).await?;
        while let Some(entry) = files.next_entry().await? {
            if entry.path().extension().map_or(true, |ext| ext != "txt") {
                continue;
            }
            tracing::debug!(
                "Text file {} has been added to task list",
                entry.path().display()
            );
            let metadata = entry.metadata().await?;
            let task = Task {
                file: entry.path(),
                size: metadata.size() as usize,
            };
            res.push(task);
        }
        res
    };

    let analysis = Arc::new(DashMap::new());
    let permits = Arc::new(Semaphore::new(args.jobs));
    let total_size: usize = all_tasks.iter().map(|t| t.size).sum();
    let analyzed_size = Arc::new(AtomicUsize::new(0));
    tracing::info!("A total of {} bytes of text to analyze", total_size);
    let mut tasks = Vec::new();
    for task in all_tasks.into_iter() {
        let analysis = Arc::clone(&analysis);
        let analyzed_size = Arc::clone(&analyzed_size);
        let permits = Arc::clone(&permits);
        let task_span = tracing::info_span!("Analysis of", file = task.file.display().to_string());
        let task = tokio::spawn(
            async move {
                tracing::debug!("Prepare to acquire a permit to start");
                let _permit = permits.acquire().await.unwrap();

                tracing::info!("Start to read file");
                let buf = fs::read(&task.file).await?;
                let len = buf.len();

                tracing::debug!("Start to analyze");
                let mut result = HashMap::new();
                for mat in ANALYZER.find_iter(&buf) {
                    let word = str::from_utf8(&buf[mat.range()]).unwrap();
                    tracing::trace!(word = %word, "Matched");
                    result
                        .entry(word.to_string())
                        .and_modify(|e| *e += 1)
                        .or_insert(1);
                    analysis
                        .entry(word.to_string())
                        .and_modify(|e| *e += 1)
                        .or_insert(1);
                }
                tracing::debug!("Finished analysis");

                let json_result = serde_json::to_vec(&result).unwrap();
                fs::write(task.file.with_extension("json"), json_result).await?;
                tracing::info!("Write result to file");

                analyzed_size.fetch_add(len, Ordering::Release);

                io::Result::Ok(())
            }
            .instrument(task_span),
        );
        tasks.push(task);
    }

    let finished = Arc::new(AtomicBool::new(false));
    let finished_flag = Arc::clone(&finished);
    let progress = tokio::spawn(async move {
        let process = ProgressBar::new(total_size as u64);
        const TEMPLATE : &str = "{spinner:.green} [{elapsed_precise}] [{wide_bar:.cyan/blue}] {bytes}/{total_bytes} ({eta})";
        let style = ProgressStyle::with_template(TEMPLATE)
            .unwrap()
            .with_key("eta", |state: &ProgressState, w: &mut dyn Write| {
                write!(w, "{:.1}s", state.eta().as_secs_f64()).unwrap()
            })
            .progress_chars("#>-");
        process.set_style(style);
        while analyzed_size.load(Ordering::Acquire) < total_size {
            if finished_flag.load(Ordering::Acquire) {
                tracing::warn!("Some tasks failed to complete");
                break;
            }
            process.set_position(analyzed_size.load(Ordering::Acquire) as u64);
            sleep(Duration::from_millis(100)).await;
        }
    });
    let results = futures::future::join_all(tasks).await;
    for res in results.into_iter().map(|r| r.unwrap()) {
        if let Err(e) = res {
            tracing::error!("failed to process a task: {}", e);
        }
    }
    finished.store(true, Ordering::Release);
    if let Err(e) = progress.await {
        tracing::error!("failed to process all tasks: {}", e);
    }

    let total_result = serde_json::to_vec(&*analysis).unwrap();
    fs::write("analysis.json", &total_result).await?;

    Ok(())
}

struct Task {
    file: PathBuf,
    size: usize,
}
